"""

The simple cases are handled except the last one in parentheses.
Run without syntax checking to check vs. the result file.

The examples are from PEP-526.

"""

class Cls:
    pass

c = Cls()
c.x: int = 0  # Annotates c.x with int.
c.y: int      # Annotates c.y with int.

d = {}
d['a']: int = 0  # Annotates d['a'] with int.
d['b']: int      # Annotates d['b'] with int.

# Note that even a parenthesized name is considered an expression, not a simple name:

(x): int      # Annotates x with int, (x) treated as expression by compiler.
(y): int = 0  # Same situation here.





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

Testfile for stripping hints.

This code doesn't really run in Python 3 unless it is stripped; it just uses
typing syntax to test stripping.

"""

from __future__ import print_function, division, absolute_import
Dict = {0:4} #  Dummy var, used in example code.
t = W = 0 # Dummies.

unicode_str = u"Protection™" # Test unicode in file.
print(unicode_str)

# Some of below tests are based on ones here:
# https://stackoverflow.com/questions/42733877/remove-type-hints-in-python-source-programmatically

e : int = lambda x,y,z: lambda: 4*(3+3)

"""Triple string"""
ggg = 44
xxx: int
xxx: list = ggg
print(xxx); print(e)

# Below line raises an exception.  Need to insert a backslash to fix it, see code note.
#tree: Dict[t
#        ] = 4
# Explicit line-continuation backslashes seem to work OK:
tree: Dict[t \
        ] = 4

# Always OK to break inside nesting level that will be retained.
chair: Dict[W] = [1,  # c1
                  2]  # c2

def default_combine_chars_fun_ZZ(elem_list: List[float], egg: int=4) -> List[str]: # REM
    pass

def in_fun():
    e=3
    e=3 # comment
    xxx: int
    xxx: int # comment
    xxx : list = ggg
    xxx : list = ggg # comment
    print(xxx)
    var : int = 4; var2 : int =34

#@decorator
def foo(bar: Dict[T, List[T]], # comment
        baz: Callable[[T], int] = 444 + 4, # comment
        **kwargs) -> List[T]: # comment
    pass

def foo(bar: Dict[T, List[T]]):
    pass

def foo(bar: Dict[T, List[T]],):
    pass

def foo(bar: Dict[T, List[T]],
        baz: Callable[[T], int] = lambda x, y: (x+3)/7,
        **kwargs) -> List[T]:
    pass

# Test NL inside parameter list, should be OK.
def colon(bar: List[
       xxx]):
    pass

# Test colon move due to NL in return type.
def colon(bar: List[xxx]) -> List[
        xxx]:
    pass

assert xxx == 44
print("done with test")




"""

Usage: strip_from_string_test.py <filename>

The file is read as a string, the string is stripped, and the output is compared
with the result read directly from the file.

"""

from __future__ import print_function, division, absolute_import

import codecs
import sys
sys.path.insert(1, "../src")

import strip_hints

if len(sys.argv) < 2:
    print("Filename required as the argument.", file=sys.stderr)
    sys.exit(1)

filename = sys.argv[1]
with codecs.open(filename, encoding='utf-8') as code_file:
    code_string = code_file.read()

stripped_string_from_file = strip_hints.strip_file_to_string(filename, only_test_for_changes=False)
stripped_string_from_string = strip_hints.strip_string_to_string(code_string, only_test_for_changes=False)

#print(stripped_string_from_string, end="", sep="")

if stripped_string_from_file != stripped_string_from_string:
    print("Error, the final stripped strings differ.")
else:
    print("The final stripped strings are identical.")




# -*- coding: utf-8 -*-
"""

Code to set up import hooks for Python 3 to automatically strip hints from
modules when they are loaded.

The import hook part of the code is based on an example from here:
   https://stackoverflow.com/questions/43571737/how-to-implement-an-import-hook-that-can-modify-the-source-code-on-the-fly-using/43573798#43573798

"""

from __future__ import print_function, division, absolute_import

import sys
import os
from importlib import invalidate_caches
from importlib.abc import SourceLoader
from importlib.machinery import FileFinder

version = sys.version_info[0]

stripper_funs = {}
registered = False

class StripHintsSourceLoader(SourceLoader):
    def __init__(self, fullname, path):
        self.fullname = fullname
        self.path = path

    def get_filename(self, fullname):
        return self.path

    def get_data(self, module_path):
        """Get the source code and modify it if necessary; `exec_module` is already
        defined."""
        assert not os.path.isdir(module_path) # Assume a file module is passed.
        canonical_module_dir_path = os.path.dirname(os.path.realpath(module_path))

        if canonical_module_dir_path in stripper_funs:
            stripper_fun_to_use = stripper_funs[canonical_module_dir_path]
            source = stripper_fun_to_use(module_path)
        else:
            with open(module_path) as f:
                source = f.read()
        return source

def install(loader_details):
    # Insert the path hook ahead of other path hooks.
    sys.path_hooks.insert(0, FileFinder.path_hook(loader_details))
    # Clear any loaders that might already be in use by the FileFinder.
    sys.path_importer_cache.clear()
    invalidate_caches()

def register_stripper_fun(calling_module_file, stripper_fun, py3_also=False):
    """The function called from a module `__init__` or from a script file to
    declare that all later imports from that directory should be processed on
    import to strip type hints.  This is based on the `realpath` of the
    directory of the module."""
    if version != 3:
        raise ImportError("Importing wrong `register_stripper_fun` for Python 2.")
    if not py3_also:
        return

    global registered
    if not registered:
        loader_details = StripHintsSourceLoader, [".py"]
        install(loader_details)
        registered = True

    canonical_dir_path = os.path.dirname(os.path.realpath(calling_module_file))
    stripper_funs[canonical_dir_path] = stripper_fun




# -*- coding: utf-8 -*-
"""

Code to set up import hooks for Python 2 to automatically strip hints from
modules when they are loaded.  Note the use of `imp` is deprecated for Python 3.

This example code was very helpful: https://github.com/aroberge/splore

Note: Consider also this `importlib` backport to Python 2:
    https://pypi.org/project/importlib2/

"""

from __future__ import print_function, division, absolute_import
import imp
import sys
import os

version = sys.version_info[0]

strip_importer_instance = None # Only install one, then modify its static attributes.

class StripHintsImporter(object):
    """Strip hints on import."""
    stripper_fun_to_use = {} # Dict of stripper funs keyed by canonical dir paths.

    def find_module(self, module_name, path=None):
        """Find the module with the given name.  Uses `path`, or `sys.path` if
        it is omitted.  Returns `(file, pathname, description)`."""
        split_module_name = module_name.split(".")
        first_name = split_module_name[0]
        rest_name = split_module_name[1:]

        full_name = first_name
        file_object, pathname, description = imp.find_module(first_name)
        self.module_info = file_object, pathname, description
        self.load_module(first_name)
        for name in rest_name:
            full_name += "." + name
            path = [pathname]
            file_object, pathname, description = imp.find_module(name, path)
            self.module_info = file_object, pathname, description
            self.load_module(full_name)

        return self

    def load_module(self, module_name):
        """Load the module."""
        if module_name in sys.modules:
            return sys.modules[module_name]

        module_path = self.module_info[1]
        file_object, pathname, description = self.module_info

        # Use regular loader unless in a dir that was registered.
        canonical_module_dir_path = os.path.dirname(os.path.realpath(module_path))
        if canonical_module_dir_path not in self.stripper_fun_to_use:
            try:
                module = imp.load_module(module_name, *self.module_info)
            finally:
                if file_object is not None:
                    file_object.close()
            return module
        else:
            stripper_fun_to_use = self.stripper_fun_to_use[canonical_module_dir_path]

        # Attempt to process the module with strip hints.
        try:
            source = stripper_fun_to_use(module_path)
            # TODO: Really should read the encoding magic comment, if there is one,
            # and encode the string in that encoding.
            if version == 2:
                source = source.encode("utf-8")
            module = imp.new_module(module_name)

            # =========================================
            # Below is based on example and other text in
            # https://www.python.org/dev/peps/pep-0302/
            mod = sys.modules.setdefault(module_path, module)
            mod.__file__ = "<%s>" % self.__class__.__name__
            mod.__loader__ = self

            ispkg = self.module_info[0] is None and self.module_info[1] is not None
            if ispkg:
                mod.__path__ = []
                mod.__package__ = module_path
            else:
                mod.__package__ = module_path.rpartition('.')[0]
            # =========================================

            sys.modules[module_name] = module
            # https://stackoverflow.com/questions/30621772/what-encoding-does-the-exec-function-assume
            exec(source, module.__dict__)
        except Exception as e:
            print("strip_hints error in loading module {0}: {1}."
                    .format(module_name, e.__class__.__name__), file=sys.stderr)
            raise # Re-raise instead of fallback to regular import.
            # module = imp.load_module(module_name, *self.module_info)
        finally:
            if file_object is not None:
                file_object.close()
        return module

def register_stripper_fun(calling_module_file, stripper_fun, py3_also=False):
    """The function called from a module `__init__` or from a script file to
    declare that all later imports from that directory should be processed on
    import to strip type hints.  This is based on the `realpath` of the
    directory of the module."""
    if version == 3:
        raise ImportError("Wrong `register_stripper_fun` imported for Python 3")

    global strip_importer_instance
    if not strip_importer_instance:
        strip_importer_instance = StripHintsImporter()
        sys.meta_path.insert(0, strip_importer_instance)

    canonical_module_dir = os.path.dirname(os.path.realpath(calling_module_file))
    strip_importer_instance.stripper_fun_to_use[canonical_module_dir] = stripper_fun




#!/usr/bin/env python
"""

Only use this script to run strip-hints from the source directory itself.
If installing from pip then a script called strip-hints will be added to
your Python scripts directory (which can be added to your PATH if it is not
already there).

The relative path of this script to the source code must be fixed, so do not
move this script.  Link to it from an alias or from a simple shell script in a
directory on your PATH to use a shorter invocation command.

"""

from __future__ import print_function, division, absolute_import
import sys
import os

bin_dir = os.path.dirname(os.path.realpath(os.path.expanduser( __file__)))
package_dir = os.path.abspath(os.path.join(bin_dir, "..", "src"))
sys.path.insert(0, package_dir)
from strip_hints.strip_hints_main import process_command_line

process_command_line()





from typing import List, NamedTuple

def foo(
  param1: int,
  param2: str = "zzz",
) -> namedtuple('outputs', [
  ('out1', str),
  ('out2', int),
]):
  pass

# x1
def foo(
  param1: int,
  param2: str = "zzz",
) -> namedtuple('outputs', [
  ('out1', str),
  ('out2', int),
]):
  # x2
  pass
# x3

egg: List[

        int

        ]

# Just before.
egg: List[

        int

        ] = 3
# Just after.

# before 2
egg: List[int] = 3
# after 2

# before 3
egg: List[int]
# after 3

# before 4
egg: List[

        int

        ]
# after 4

x = 3
egg: List[int]
x = 3

egg

def f():
    egg: List[int]

    pass

def f():
    # f 1
    egg: List[int]
    # f 2

    pass





def simp(x: int, y: bool) -> bool:
    def q(x: int=4):
        x: int=4

egg : int

def foo(bar: Dict[T, List[T]], egg=lambda x,y: 4*(3+3),
        egg2: int= lambda: (4),
        baz: Callable[[T], int] = 444 + 4,
        **kwargs) -> List[T]:
    pass

#
# Below test some cases that might get confused with an annotated variable or assignment.
#

m = 2
string = 4
if m == ")": string += m
else: string = m + string

if m == ")": \
    string += m
else: \
    string = m + string

if m == ")": \
    string += m
else: \
    string = m + string

if m == ")": # comment
    string += m
else: # comment
    string = m + string





class Tst:
    var1: int = 99

    class Nested:
        x: int = 4

    def method(self, fff: list):
        self.xyz :int = 100
        qwerty : dict = {}
        self.nested.x : int = 3
        self.nested.x : int

    @property
    def size():
        size: int = -1
        return size




#!/usr/bin/env python3
"""

Try transforming on import by setting the import hook.

"""

from __future__ import print_function, division, absolute_import

import strip_hints

strip_hints.strip_on_import(__file__, py3_also=True)

import testfile_strip_hints

from simple_test import m
assert m == 2

# Some general import tests below.
import sys
import xml.etree.ElementTree
print(sys.modules.keys())
assert "xml" in sys.modules
assert "xml.etree" in sys.modules
assert "xml.etree.ElementTree" in sys.modules
import xml # Make sure other imports work...
print("\ndir(xml)", dir(xml))
import xml.parsers # Make sure other imports work...
x = xml.parsers
from xml import parsers
assert "xml.parsers" in sys.modules
print("\ndir(xml.parsers)", dir(xml.parsers))
#import xml.etree
#print("\ndir(xml.etree)", xml.etree.__dict__.keys())
from xml.etree import ElementTree


